import pandas
from PrettyPrint.output import Printer
from PrettyPrint.figures.Separator import separator
from PrettyPrint.format import *

class Table:
    def __init__(self, data=None, columns=None, printer=None, cell_num=None, col_width=None, cell_fmt=None, frame_fmt=None):
        # check for printer arg
        if isinstance(printer, Printer):
            self._printer = printer
        elif printer is None:
            self._printer = Printer()
        else:
            raise TypeError('printer must be a Printer or None')

        # check for dataframe
        if isinstance(data, pandas.DataFrame):
            self.df = data
            self.cols = len(self.df.columns)
        elif data is None:
            self.df = pandas.DataFrame(columns=columns, index=None)
            self.cols = len(self.df.columns)
        else:
            raise TypeError('data has to be a pandas.DataFrame or None')

        # create cell fmt list
        if cell_fmt is None:
            self.cell_fmt = ['']*self.cols
        elif isinstance(cell_fmt, PPFormat):
            self.cell_fmt = [str(cell_fmt)]*self.cols
        elif isinstance(cell_fmt, list) and all(isinstance(elem, PPFormat) for elem in cell_fmt):
            if len(cell_fmt) == self.cols:
                self.cell_fmt = cell_fmt
            else:
                raise ValueError('cell_fmt list must be same length as cols')
        else:
            raise TypeError('cell_fmt must be None, PPFormat or list of PPFormat')

        # check for df size
        if len(self.df.columns) > 10:
            raise UserWarning('Column number is large and may not fit in the console window.')

        # check for frame format
        if frame_fmt is None:
            self._vertical_sep = '|'
            self.frame_fmt = Default()
        elif isinstance(frame_fmt, PPFormat):
            self._vertical_sep = str(frame_fmt)+'|'+"\033[0m"
            self.frame_fmt = frame_fmt
        else:
            raise TypeError('frame_fmt needs to be a PPFormat or None')

        # build format string
        self._string = self._vertical_sep
        self._string += ('{}{}' + self._vertical_sep) * self.cols
        if type(cell_num) is str:
            self._cell_num = [cell_num]*self.cols
        elif cell_num is None:
            self._cell_num = ['']*self.cols
        elif isinstance(cell_num, list) and all(isinstance(elem, str) for elem in cell_num):
            if len(cell_num) == self.cols:
                self._cell_num = cell_num
            else:
                raise ValueError('cell_num list must be same length as cols')
        else:
            raise TypeError('cell_num needs to be None, a single format or a list of formats')

        # compute width
        if col_width is None:
            self._col_width = [len(elem) for elem in self.df.columns]
        elif type(col_width) is int:
            self._col_width = [col_width for elem in self.df.columns]
        elif isinstance(col_width, list) and all(isinstance(elem, int) for elem in col_width):
            if len(col_width) == self.cols:
                self._col_width = col_width
            else:
                raise ValueError('col_width list must be same length as cols')
        else:
            raise TypeError('col_width needs to be a list of integers, an integer or none')

        self._table_width = sum(self._col_width)+1+self.cols

        if data is not None:
            self._print_table()
        else:
            self._print_header()


    def add_row(self, series):
        # add new row to df, call print_row after to print to console
        if len(series) == self.cols:
            self.df.loc[len(self.df.index)] = series
            self._print_row(-1)
        else:
            raise ValueError('Series and table of unequal length.')

    def _resize_cell(self, idx, content):
        if self._col_width[idx] < len(content):
            return content[:self._col_width[idx]]
        elif self._col_width[idx] > len(content):
            padding = self._col_width[idx] - len(content)
            return ' ' * padding + content
        else:
            return content

    def _parse_row(self, idx): # uses pd series objects
        series = self.df.iloc[idx]
        args = []
        for i in range(len(series)):
            args.append(self.cell_fmt[i])
            args.append(self._resize_cell(i, ('{'+self._cell_num[i]+'}').format(series.iloc[i])))
        row_string = self._string.format(*args)
        return row_string

    def _parse_header(self): # parses the header to the specified format
        header_string = self._vertical_sep+('{}{}'+self._vertical_sep)*self.cols
        args = []
        for i in range(self.cols):
            args.append(self.cell_fmt[i])
            args.append(self._resize_cell(i, self.df.columns[i]))
        row_string = header_string.format(*args)
        return row_string

    def _print_sep(self):
        self._printer(separator(self._table_width), fmt=self.frame_fmt)

    def _print_header(self):
        self._print_sep()
        head_str = self._parse_header()  # build row string
        self._printer(head_str)
        self._print_sep()

    def _print_row(self, idx):
        row_str = self._parse_row(idx) # build row string
        self._printer(row_str)
        self._print_sep()

    def _print_table(self):
        self._print_header()
        for i in range(len(self.df.index)):
            self._print_row(i)





import sys
import time

class ProgressBar:
    '''
    The ProgressBar class, meant to be init before a finite loop with known iteration count,
    to show the progress of the loop for loops
    Consider that there should be no other prints during the use of this pattern, because that will fuck up the print with newlines
    '''
    def __init__(self, underlying_iterable, bins=10, symbol='='):
        # dtype checking lessgo
        try:
            iter(underlying_iterable)
        except TypeError:
            raise TypeError('Progress Bar expects an iterable but got {} instead'.format(type(underlying_iterable)))
        if type(bins) is not int:
            raise TypeError('bins must be a positive integer between 1 and 100')
        if bins <= 0:
            raise ValueError('bins must be a positive integer between 1 and 100')
        if bins > 100:
            raise ValueError('bins must be a positive integer between 1 and 100')
        if type(symbol) is not str:
            raise TypeError('symbol must be a char but got {}'.format(type(symbol)))

        # properties for utils
        self.underlying_iterable = iter(underlying_iterable) # get the iterator of the underlying itrable
        self._length = sum(1 for e in underlying_iterable) # compute the length of the underlying iterable
        self._step = round(self._length/bins) # compute the stepsize (=amount of iterations to fill one bin
        self._sym = symbol # store the symbol with which the bar is printed
        self._index = 0 # count iterations for each step
        self._progress = 0 # track progress
        self._bins = bins # the number of bins
        self._perc_increment = 100/bins # the percentual increments for every bin
        self._formatter = "[%-"+str(bins)+"s] %d%% | AVG loop time: %.2f sec| time passed: %.2f sec | estimated time left: %.2f sec" # the formatter string to print each update
        # another error handling
        if bins > self._length:
            raise ValueError('the bar cant have more bins than iterations in the underlying iterable')

        # properties for timing
        self._start_time = time.time() # saves the start time
        self._time_elapsed = 0
        self._avg_iter_time = None # saves the time an iteration takes to complete on average
        self._iter = 0 # saves the current iter count


    def __next__(self):
        self._increment()
        self._iter += 1
        return self.underlying_iterable.__next__()

    def __iter__(self):
        return self

    def __len__(self):
        return self.underlying_iterable.__len__()

    def _increment(self):
        '''
        increments the index, if step size is reached updates the printout and resets index
        :return:
        '''
        self._index += 1
        if self._index == self._step:
            self._update()
            self._index = 0
        elif self._iter == 1:
            self._update()

    def _update(self):
        '''
        Adds a tick to the progress bar. will simply stop updating if the loop goes out of bounds of the previously stated max iteration count
        :return:
        '''
        if self._progress <= self._bins:
            sys.stdout.write('\r') # resets the cursor

            self._time_elapsed = time.time()-self._start_time # update elapsed time
            self._avg_iter_time = self._time_elapsed/self._iter # update average time per loop
            self._est_time_left = self._avg_iter_time*(self._length - self._iter) # update time left

            sys.stdout.write((self._formatter) % (
                self._sym * self._progress,
                self._perc_increment * self._progress,
                self._avg_iter_time,
                self._time_elapsed,
                self._est_time_left
            )) # writes the formatted string

            sys.stdout.flush() # flush buffer to print text immediately
            self._progress += 1





from PrettyPrint import ansi


class ColourText:
    def __init__(self, colour='default', option='default'):
        if isinstance(colour, str):
            if colour in ansi._COLOURS:
                if option == 'default':
                    self.fmt = f"\033[3{ansi._COLOURS[colour]}m"
                elif option == 'bright':
                    self.fmt = f"\033[9{ansi._COLOURS[colour]}m"
                else:
                    raise ValueError(f"Selected option doesnt exist, must be default or bright, got {option}")
            else:
                raise ValueError(f"Selected colour is not supported: {colour}")
        elif isinstance(colour, list) and all(isinstance(elem, int) for elem in colour) and isinstance(option, str):
            if option == 'default':
                self.fmt = f"\033[38;2;{colour[0]};{colour[1]};{colour[2]}m"
            elif option == 'bright':
                self.fmt = f"\033[98;2;{colour[0]};{colour[1]};{colour[2]}m"
            else:
                raise ValueError(f"Selected option doesnt exist, must be default or bright, got {option}")

            #self.fmt = f"\033[38;2;{colour[0]};{colour[1]};{colour[2]}m"
        else:
            raise TypeError('Passed colour is not a supported type! Supported types are string and list of int, got {} instead'.format(type(colour)))

    def __call__(self):
        return self.fmt


class ColourBackground:
    def __init__(self, colour='default', option='default'):
        if isinstance(colour, str):
            if colour in ansi._COLOURS:
                if option == 'default':
                    self.fmt = f"\033[4{ansi._COLOURS[colour]}m"
                elif option == 'bright':
                    self.fmt = f"\033[10{ansi._COLOURS[colour]}m"
                else:
                    raise ValueError(f"Selected option doesnt exist, must be default or bright, got {option}")
            else:
                raise ValueError(f"Selected colour is not supported: {colour}")
        elif isinstance(colour, list) and all(isinstance(elem, int) for elem in colour) and isinstance(option, str):
            if option == 'default':
                self.fmt = f"\033[48;2;{colour[0]};{colour[1]};{colour[2]}m"
            elif option == 'bright':
                self.fmt = f"\033[108;2;{colour[0]};{colour[1]};{colour[2]}m"
            else:
                raise ValueError(f"Selected option doesnt exist, must be default or bright, got {option}")

        else:
            raise TypeError(
                'Passed colour is not a supported type! Supported types are string and list of int, got {} instead'.format(
                    type(colour)))

    def __call__(self):
        return self.fmt

class Effect:
    def __init__(self, effect):
        if isinstance(effect, str):
            if effect in ansi._EFFECTS:
                self.fmt = ansi._EFFECTS[effect]
            else:
                raise ValueError(f"Selected effect is not supported")
        else:
            raise TypeError(f"Passed effect is not a supported type, must be String but got {type(effect)}")

    def __call__(self):
        return self.fmt

class Font:
    def __init__(self, font):
        if isinstance(font, str):
            if font in ansi._FONTS:
                self.fmt = ansi._FONTS[font]
            else:
                raise ValueError(f"Selected font is not supported")
        else:
            raise TypeError(f"Passed font is not a supported type, must be String but got {type(font)}")

    def __call__(self):
        return self.fmt

class PPFormat:
    def __init__(self, options='default'):
        if options == 'default':
            self.format = ''
        else:
            self.compose(options)

    def compose(self, structure):
        '''
        Concatenates all the subsequences to make a big combined sequence for formatting
        :param structure: a list of callables with objects from ansi_util
        :return:
        '''
        if self._is_list_of_callables(structure):
            fmt = ''
            for element in structure:
                fmt += element()
            self.format = fmt
        else:
            self.format = ''
            raise TypeError('The passed variable is not a list of callables.')


    def _is_list_of_callables(self, variable) -> bool:
        '''
        Checks if the passed argument is a list of callables
        :param variable: the argument to check
        :return: bool
        '''
        if isinstance(variable, list) and all(callable(item) for item in variable):
            return True
        return False

    def __str__(self):
        return self.format

class Warning(PPFormat):
    def __init__(self):
        super().__init__([ColourText('yellow'),Effect('underlined'),Effect('bold')])

class Default(PPFormat):
    def __init__(self):
        super().__init__([Effect('reset')])

class Error(PPFormat):
    def __init__(self):
        super().__init__([ColourText('red'),Effect('underlined'),Effect('bold')])

class Success(PPFormat):
    def __init__(self):
        super().__init__([ColourText('green'),Effect('underlined'),Effect('bold')])


from PrettyPrint.figures.ProgressBar import ProgressBar
from PrettyPrint.figures.RunningIndicator import RunningIndicator
from PrettyPrint.figures.Table import Table
from PrettyPrint.figures.Separator import separator



import sys
import time

class RunningIndicator:
    '''
    This class provides an indication for running while loops. Gives information about time passed but cant make
    estimations about time left and progress done
    just call this object at the beginning of each loop
    '''
    def __init__(self, mode='dots', update_frq = 1):
        self._state = 0 # which state of the charset gets printed at update time
        self._formatter = "Running %s | AVG loop time: %.2f sec| time passed: %.2f sec"  # the formatter string to print each update
        self._update_frq = update_frq # the amount of loops to pass before the output gets updated

        # select the given charset
        if mode == 'dots':
            self._chars = ['.  ', '.. ', '...']
        elif mode == 'rotate':
            self._chars = ['- ', '\ ', '/ ']
        else:
            raise ValueError('{} is not a valid mode'.format(mode))

        # properties for timing
        self._start_time = time.time()  # saves the start time
        self._time_elapsed = 0
        self._iter = 0

    def __call__(self):
        self._iter += 1 # count passed iters
        if self._iter == 1:
            self._time_elapsed = time.time() - self._start_time  # update elapsed time
            self._avg_iter_time = self._time_elapsed / self._iter  # update average time per loop

            sys.stdout.write('\r')  # resets the cursor
            sys.stdout.write(
                self._formatter % (self._chars[self._state], self._avg_iter_time, self._time_elapsed))  # prints the msg
            sys.stdout.flush()

            # update internal state
            self._state += 1

        if self._iter % self._update_frq == 0: # check if frq time is met
            self._time_elapsed = time.time() - self._start_time  # update elapsed time
            self._avg_iter_time = self._time_elapsed / self._iter  # update average time per loop

            sys.stdout.write('\r')  # resets the cursor
            sys.stdout.write(self._formatter % (self._chars[self._state], self._avg_iter_time, self._time_elapsed)) # prints the msg
            sys.stdout.flush()

            # update internal state
            self._state += 1
            if self._state == 3:
                self._state = 0



from setuptools import setup, find_packages

VERSION = '1.0.2'
DESCRIPTION = 'A simple package that contains a set of useful printing options'
LONG_DESCRIPTION = 'This Package uses the ANSI escape sequences to format console printouts as well as animations such as a progress bar for use in the console. Furthermore options to automatically create logfiles of each statement printed. See the GitHub repo for more info: https://github.com/Bangulli/PrettyPrint'

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="PrettyPrintout",
    version=VERSION,
    author="Lorenz Kuhn",
    author_email="<lorenz.achim.kuhn@gmail.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],  # add any additional packages that
    # needs to be installed along with your package. Eg: 'caer'

    keywords=['python', 'first package', 'console formatting', 'output', 'logging', 'printing'],

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import sys

import PrettyPrint.format as format
import time


class Printer:
    def __init__(self, log_type=None, timestamps=False, log_prefix=''):
        '''
        Initialize the Outputter. On call takes a text and prints it to console with the specified formatting
        Also logs everything to a file if log_type is {html, txt}
        :param log_type: the type of logfile created, if None, no file is created. if 'html' formatted html, if 'txt' plain text txt
        :param timestamps: controls whether to print timestamps before each output
        '''
        if log_type in ['html', 'txt']:
            self.log_type = log_type
            filename = f"""{time.strftime("%Y%m%d-%H%M")}_{log_prefix}_PrettyPrint_Autolog.{self.log_type}"""
            self.log = open(filename, 'w')
            if log_type == 'html':
                html_init_str = '<!DOCTYPE html>\n<html>\n  <head>\n        <title>'+filename+'</title>\n   </head>\n   <body>\n'
                self.log.write(html_init_str)
        else:
            self.log = None

        self.timestamps = timestamps

    def __call__(self, msg, fmt=None):
        '''
        Prints a formatted message to the console and logs it to a file if log_type is {html, txt}
        :param msg: String The text to be printed
        :param fmt: String The format to print with, constructed using Builder.Compose() with the ansi_util callables.
        :return:
        '''
        if isinstance(fmt, format.PPFormat):
            formatter = str(fmt)
        elif fmt is None:
            formatter = "\033[0m"
        else:
            raise ValueError(
                f"fmt is not a suppoerted datastructure, expected format.PPFormat or None, got {str(type(fmt))} instead")

        if self.timestamps:
            sys.stdout.write(f"""{time.strftime("%Y-%m-%d-%H:%M:%S - ")}{formatter}{msg}\033[0m\n""")
        else:
            sys.stdout.write(f"{formatter}{msg}\033[0m\n")

        if self.log is not None:
            if self.log_type == 'html':
                #self.log.write("""      <p style="color: {}; font-family: 'Liberation Sans',sans-serif">{}</p>\n""".format(colour, msg))
                self.log.write(msg + '\n')

            elif self.log_type == 'txt':
                self.log.write(msg+'\n')


    def __del__(self):
        '''
        Called on delete to close the file and finish html if log_type is html
        :return:
        '''
        if self.log is not None:
            if self.log_type == 'html':
                html_closing_str = '    </body>\n</html>'
                self.log.write(html_closing_str)
            self.log.close()


    def tagged_print(self, tag, msg, tag_fmt, msg_fmt=format.Default()):
        if isinstance(tag_fmt, format.PPFormat):
            tag_fmt = str(tag_fmt)
        else:
            raise TypeError(f"Unbrecognized type for tag format: {type(tag_fmt)}")
        if isinstance(msg_fmt, format.PPFormat):
            msg_fmt = str(msg_fmt)
        else:
            raise TypeError(f"Unbrecognized type for message format: {type(msg_fmt)}")

        if self.timestamps: #
            self.__call__(f"""{time.strftime("%Y-%m-%d-%H:%M:%S - ")}{tag_fmt}{tag}\033[0m:{msg_fmt} {msg}\033[0m""")
        else:
            self.__call__(f"{tag_fmt}{tag}\033[0m:{msg_fmt} {msg}\033[0m")


    def warning(self, msg):
        self.tagged_print('WARNING', msg, format.Warning(), format.Default())

    def error(self, msg):
        self.tagged_print('ERROR', msg, format.Error(), format.Default())

    def success(self, msg):
        self.tagged_print('SUCCESS', msg, format.Success(), format.Default())

    def fail(self, msg):
        self.tagged_print('FAIL', msg, format.Error(), format.Default())

